#include <avr/interrupt.h>
#include <avr/io.h>
#include <util/delay.h>

#include "connect4.h"

volatile uint8_t board[ROWS_NR][COLS_NR] = {NA};
volatile int8_t curr_col = 3, curr_color = BRED;
volatile int8_t curr_row;
volatile int8_t blink_on = BLINK_ON;
volatile int8_t blink_off = BLINK_OFF;
volatile int8_t blink = ON;
volatile int8_t use_buttons = 1;
volatile int8_t operation;
volatile int8_t won;
volatile int8_t buzzer_cnt = BUZZER_CNT;
volatile int8_t buzzer_time = 100;
volatile int8_t first = 1;

ISR(TIMER1_COMPA_vect)
{
	cli();
	if (operation == PLACE_OP) {
		curr_row++;
		if (curr_row != ROWS_NR && is_empty(curr_row, curr_col)) {
			board[curr_row][curr_col] = board[curr_row - 1][curr_col];
			board[curr_row - 1][curr_col] = NA;
		}
		else {
			timer1_uninit();
			if (game_over()) {
				won = 1;
				win(curr_color);
				timer3_init(buzzer_time);
			}
			else {
				change_turn();
				use_buttons = 1;
			}
		}
	}
	else {
		int i;
		for (i = 0; i < ROWS_NR; i++)
			if (!is_empty(i, curr_col))
				break;
		if (i == 0) {
			use_buttons = 1;
			timer1_uninit();
			board[i][curr_col] = NA;
			change_turn();
		}
		else {
			board[i - 1][curr_col] = board[i][curr_col];
			board[i][curr_col] = NA;
		}
	}
	sei();
}

ISR(TIMER1_COMPB_vect)
{
	cli();
	if (won) {
		PIN_BUZZER |= (1 << POS_BUZZER);
	}
	else {
		use_buttons = 1;
		timer2_uninit();
	}
	sei();
}

void turn_on_led(int row, int col, int color)
{
	if (color == NA)
		return;
	if (color == BRED || color == BYEL) {
		color -= 3;
		if (blink == ON) {
			blink_on--;
			if (blink_on <= 0) {
				blink_on = BLINK_ON;
				blink = OFF;
				return;
			}
		}
		else {
			blink_off--;
			if (blink_off > 0)
				return;
			blink_off = BLINK_OFF;
			blink = ON;
		}
	}
	(*row_ports[color][row]) &= ~(1 << row_pins[color][row]);
	(*col_ports[color][col]) |= (1 << col_pins[color][col]);
}

void turn_off_led(int row, int col, int color)
{
	if (color == NA)
		return;
	if (color == BRED || color == BYEL)
		color -= 3;
	(*row_ports[color][row]) |= (1 << row_pins[color][row]);
	(*col_ports[color][col]) &= ~(1 << col_pins[color][col]);
}

void timer1_init()
{
	ICR1 = 2000;
	OCR1A = ICR1;
	TIMSK1 = (1 << OCIE1A);
	TCCR1B = (5 << CS10) | (1 << WGM12) | (1 << WGM13);
}

void timer1_uninit()
{
	TIMSK1 = 0;
	TCCR1B = 0;
}

void timer2_init()
{
	ICR1 = 4000;
	OCR1B = ICR1;
	TIMSK1 = (1 << OCIE1B);
	TCCR1B = (5 << CS10) | (1 << WGM12) | (1 << WGM13);
}

void timer2_uninit()
{
	TIMSK1 = 0;
	TCCR1B = 0;
}

void timer3_init(int time)
{
	ICR1 = time;
	OCR1B = ICR1;
	TIMSK1 = (1 << OCIE1B);
	TCCR1B = (5 << CS10) | (1 << WGM12) | (1 << WGM13);
}

void timer3_uninit()
{
	TIMSK1 = 0;
	TCCR1B = 0;
}

void init()
{
	int i, j;
	sei();
	// set row pins as output and set them to 1 (turned off)
	for (i = 0; i < COLORS_NR; i++) {
		for (j = 0; j < ROWS_NR; j++) {
			(*row_ports[i][j]) |= (1 << row_pins[i][j]);
			(*row_ddrs[i][j]) |= (1 << row_pins[i][j]);
		}
	}
	// set column pins as output
	for (i = 0; i < COLORS_NR; i++) {
		for (j = 0; j < COLS_NR; j++) {
			(*col_ddrs[i][j]) |= (1 << col_pins[i][j]);
		}
	}
	// set buttons as input
	DDR_DOWN &= ~(1 << POS_DOWN);
	DDR_LEFT &= ~(1 << POS_LEFT);
	DDR_RIGHT &= ~(1 << POS_RIGHT);
	DDR_UP &= ~(1 << POS_UP);

	DDR_BUZZER |= (1 << POS_BUZZER);
	// activate pull-up resistor for buttons
	PORT_DOWN |= (1 << POS_DOWN);
	PORT_LEFT |= (1 << POS_LEFT);
	PORT_RIGHT |= (1 << POS_RIGHT);
	PORT_UP |= (1 << POS_UP);

	PORT_BUZZER &= ~(1 << POS_BUZZER);
}

void display_board()
{
	int i;
	for (i = 0; i < COLS_NR; i++) {
		turn_on_led(0, i, board[0][i]);
		turn_on_led(1, i, board[1][i]);
		turn_on_led(2, i, board[2][i]);
		_delay_ms(ROW_DELAY);
		turn_off_led(0, i, RED);
		turn_off_led(0, i, YEL);
		turn_off_led(1, i, RED);
		turn_off_led(1, i, YEL);
		turn_off_led(2, i, RED);
		turn_off_led(2, i, YEL);
		turn_on_led(3, i, board[3][i]);
		turn_on_led(4, i, board[4][i]);
		turn_on_led(5, i, board[5][i]);
		_delay_ms(ROW_DELAY);
		turn_off_led(3, i, RED);
		turn_off_led(3, i, YEL);
		turn_off_led(4, i, RED);
		turn_off_led(4, i, YEL);
		turn_off_led(5, i, RED);
		turn_off_led(5, i, YEL);
	}
}

int is_empty(int row, int col)
{
	return board[row][col] != RED && board[row][col] != YEL;
}

void change_turn()
{
	if (curr_color == RED || curr_color == BRED)
		curr_color = BYEL;
	else
		curr_color = BRED;
	if (is_empty(0, curr_col))
		board[0][curr_col] = curr_color;
}

void place_piece()
{
	if (!is_empty(0, curr_col))
		return;
	if (first == 1)
		first = 2;
	curr_color -= 3;
	curr_row = 0;
	board[curr_row][curr_col] = curr_color;
	operation = PLACE_OP;
	use_buttons = 0;
	timer1_init();
}

void remove_piece()
{
	if (is_empty(ROWS_NR - 1, curr_col))
		return;
	operation = REMOVE_OP;
	use_buttons = 0;
	timer1_init();
}

void win(int color)
{
	int i, j;
	for (i = 0; i < ROWS_NR; i++)
		for (j = 0; j < COLS_NR; j++)
			board[i][j] = color + 3;
}

int check_row(int row)
{
	if (is_empty(row, 3))
		return 0;
	int win_color = board[row][3];
	int cnt = 0;
	int i;
	for (i = 0; i < COLS_NR; i++) {
		if (board[row][i] == win_color)
			cnt++;
		else
			cnt = 0;
		if (cnt == WIN_NR)
			return 1;
	}
	return 0;
}

int check_rows()
{
	int i;
	for (i = 0; i < ROWS_NR; i++)
		if (check_row(i))
			break;
	if (i != ROWS_NR)
		return 1;
	return 0;
}

int check_col(int col)
{
	if (is_empty(3, col))
		return 0;
	int win_color = board[3][col];
	int cnt = 0;
	int i;
	for (i = 0; i < ROWS_NR; i++) {
		if (board[i][col] == win_color)
			cnt++;
		else
			cnt = 0;
		if (cnt == WIN_NR)
			return 1;
	}
	return 0;
}

int check_cols()
{
	int i;
	for (i = 0; i < COLS_NR; i++)
		if (check_col(i))
			break;
	if (i != COLS_NR)
		return 1;
	return 0;
}

int check_diag_right(int row, int col)
{
	if (is_empty(row, col))
		return 0;
	if (board[row][col] == board[row + 1][col + 1] &&
			board[row][col] == board[row + 2][col + 2] &&
			board[row][col] == board[row + 3][col + 3])
		return 1;
	return 0;
}

int check_diag_left(int row, int col)
{
	if (is_empty(row, col))
		return 0;
	if (board[row][col] == board[row + 1][col - 1] &&
			board[row][col] == board[row + 2][col - 2] &&
			board[row][col] == board[row + 3][col - 3])
		return 1;
	return 0;
}

int check_diags()
{
	int i, j;
	for (i = 0; i <= 2; i++) {
		for (j = 0; j <= 3; j++)
			if (check_diag_right(i, j))
				return 1;
		for (j = 3; j < COLS_NR; j++)
			if (check_diag_left(i, j))
				return 1;
	}
	return 0;
}

int game_over()
{
	if (check_rows() || check_cols() || check_diags())
		return 1;
	return 0;
}

void loop()
{
	int i, j;
	for (i = 0; i < ROWS_NR; i++)
		for (j = 0; j < COLS_NR; j++)
			board[i][j] = NA;
	board[0][curr_col] = curr_color;

	while(1) {
		if (use_buttons) {
			if ((PIN_DOWN & (1 << POS_DOWN)) == 0)
				place_piece();
			if ((PIN_LEFT & (1 << POS_LEFT)) == 0) {
				if (is_empty(0, curr_col))
					board[0][curr_col] = NA;
				if (first == 2)
					first = 0;
				else
					curr_col--;
				if (curr_col < 0)
					curr_col = COLS_NR - 1;
				if (is_empty(0, curr_col))
					board[0][curr_col] = curr_color;
				use_buttons = 0;
				timer2_init();
			}
			if ((PIN_RIGHT & (1 << POS_RIGHT)) == 0) {
				if (is_empty(0, curr_col))
					board[0][curr_col] = NA;
				if (first == 2)
					first = 0;
				else
					curr_col++;
				if (curr_col >= COLS_NR)
					curr_col = 0;
				if (is_empty(0, curr_col))
					board[0][curr_col] = curr_color;
				use_buttons = 0;
				timer2_init();
			}
			if ((PIN_UP & (1 << POS_UP)) == 0)
				remove_piece();
		}
		display_board();
	}
}

int main(void)
{
	init();
	loop();

	return 0;
}
